import random
import unittest

from plugin.occupancy_processor import FreeSeatsProcessor, FillRateProcessor
from plugin.occupancy_processor.occupancy_processor import OccupancyProcessor
from src import EventListener, Broker, Event


class OccupancyProcessorTestListener(EventListener):
    def __init__(self):
        super().__init__()
        self.received = []
        self.set_message_type_callback('processed_occupancy', lambda x: self.received.append(x))


class OccupancyProcessorTests(unittest.TestCase):
    def test_occupancy_processor(self):
        test_occupancy_1 = {
            '1': True,
            '2': False,
            '3': True,
            '4': False,
            '5': True,
            '6': False,
        }

        # Create and subscribe listeners
        broker = Broker()
        broker.clear_listeners()
        occupancy_processor = OccupancyProcessor('processed_occupancy', FreeSeatsProcessor())
        occupancy_processor.add_sub_processor(FillRateProcessor())
        test_listener = OccupancyProcessorTestListener()
        broker.subscribe(test_listener, 'processed_occupancy')

        # Publish occupancy
        broker.publish(Event('occupancy', test_occupancy_1))

        # Check result
        self.assertEqual(test_listener.received[0].processed_data['fill_rate'], 0.5)
        self.assertEqual(test_listener.received[0].processed_data['free_seats'], 3)
        self.assertEqual(test_listener.received[0].seat_map, test_occupancy_1)

    def test_occupancy_processor_multiple(self):
        test_occupancy_1 = {str(i): False for i in range(1, 7)}
        test_occupancy_2 = {str(i): True for i in range(1, 7)}

        broker = Broker()
        broker.clear_listeners()

        # create random seat occupancy
        random.seed(64)
        nr_of_seats = random.randint(5, 100)
        nr_occupied = random.randint(0, nr_of_seats)
        occupancy_list = [True for _ in range(nr_occupied)]
        occupancy_list.extend([False for _ in range(nr_of_seats - nr_occupied)])
        random.shuffle(occupancy_list)
        test_occupancy_3 = {str(i + 1): occupancy_list[i] for i in range(nr_of_seats)}

        # Create listener and subscribe
        # occupancy_processor already created in previous tests
        occupancy_processor = OccupancyProcessor('processed_occupancy', FreeSeatsProcessor())
        occupancy_processor.add_sub_processor(FillRateProcessor())
        test_listener = OccupancyProcessorTestListener()
        broker.subscribe(test_listener, 'processed_occupancy')

        # Publish occupancies
        broker.publish(Event('occupancy', test_occupancy_1))
        broker.publish(Event('occupancy', test_occupancy_2))
        broker.publish(Event('occupancy', test_occupancy_3))

        # Check if each processed instance is correct
        for i, fill_rate, free_seats, test_occupancy in [(0, 0, 6, test_occupancy_1), (1, 1, 0, test_occupancy_2), (
                2, nr_occupied / nr_of_seats, nr_of_seats - nr_occupied, test_occupancy_3)]:
            self.assertEqual(test_listener.received[i].processed_data['fill_rate'], fill_rate)
            self.assertEqual(test_listener.received[i].processed_data['free_seats'], free_seats)
            self.assertEqual(test_listener.received[i].seat_map, test_occupancy)



from src import Broker


class InternalAPI:
    """
    This component handles all network requests made by clients (staff, travellers, ...) aboard the train.
    This component receives the most recent seat occupancy data from the train unit data aggregator and stores this
    data such that incoming received network requests can be fulfilled using only this locally stored data.
    """

    def __init__(self):
        self.occupancy_data_map = {}

    def receive_occupancy_data(self, occupancy_data):
        """
        Receive and store the updated seat occupancy data from the train unit data aggregator.
        """

        id = occupancy_data.carriage_id
        self.occupancy_data_map[id] = occupancy_data.to_API_object()

    def destroy(self):
        """Handles all destroy logic which should for example close reactor listeners."""
        Broker().unsubscribe(self)



"""Module that contains test objects that allow developers to write tests for network logic"""

from abc import ABC

from twisted.internet.protocol import DatagramProtocol, Protocol
from twisted.test.proto_helpers import StringTransport

from src.network_daemon import TCPServer, TCPClient
from src.network_daemon.multicast_network_objects import NetworkServer, NetworkClient


class MockingTransport(StringTransport):
    """Mocking transport which takes strings"""


class MockingNetworkObject(DatagramProtocol):
    """Mocking network object which uses the MockingTransport"""

    def __init__(self):
        super(DatagramProtocol, self).__init__()
        self.transport = MockingTransport()


class MockingNetworkServer(MockingNetworkObject, NetworkServer):
    """Mocking network server which uses the MockingTransport"""


class MockingNetworkClient(MockingNetworkObject, NetworkClient):
    """Mocking network client which uses the MockingTransport and implements the send_message logic"""

    def send_message(self, message):
        if isinstance(message, str):
            self.transport.write(str.encode(message))
        elif isinstance(message, bytes):
            self.transport.write(message)
        else:
            raise ValueError('Invalid message, should be bytes or str.')


class MockingNetworkProtocol(Protocol):
    """Mocking class for protocols"""

    def __init__(self):
        self.transport = MockingTransport()


class MockingNetworkTCPServer(MockingNetworkProtocol, TCPServer, ABC):
    """Mocking TCP server"""

    def __init__(self):
        MockingNetworkProtocol.__init__(self)


class MockingNetworkTCPClient(MockingNetworkProtocol, TCPClient):
    """Mocking TCP client"""

    def __init__(self):
        MockingNetworkProtocol.__init__(self)



from twisted.internet.protocol import DatagramProtocol


class MulticastObject(DatagramProtocol):
    """A multicast twisted object which uses a multicast group and a port"""

    def __init__(self, group, port):
        self.group = group
        self.port = port

    def startProtocol(self):
        """Start the protocol by joining the given multicast group"""
        self.transport.joinGroup(self.group)

    def stopProtocol(self):
        if self.transport is not None:
            self.transport.loseConnection()


class NetworkServer:
    """Network server interface"""

    def datagramReceived(self, datagram, address):
        """Do something when a datagram has been received. Override this."""
        raise NotImplementedError()


class MulticastNetworkServer(MulticastObject, NetworkServer):
    """Multicast server which listens to a multicast group"""
    pass


class NetworkClient(DatagramProtocol):
    def send_message(self, message):
        if isinstance(message, str):
            self.transport.write(str.encode(message), (self.group, self.port))
        elif isinstance(message, bytes):
            self.transport.write(message, (self.group, self.port))
        else:
            raise ValueError('Invalid message, should be bytes or str.')


class MulticastNetworkClient(MulticastObject, NetworkClient):
    """Multicast client which is able to send messages"""
    pass



import unittest
import uuid

from twisted.internet import reactor

from src import ElectionComponent, NetworkElectionComponent
from src import STATES
from src import TudaElectionComponent
from src import MockingNetworkClient, MockingNetworkServer


class _TestElectionComponent(ElectionComponent):
    def __init__(self, id):
        super().__init__(id)

    def on_elected(self, is_leader):
        pass

    def __str__(self):
        return '{} - [{}]'.format(self._id, self._others)


class TestElectionComponent(unittest.TestCase):
    def _update_others(self, election_components):
        for i, election_component in enumerate(election_components):
            election_component._others = election_components[:i] + election_components[i + 1:]

    def test_election(self):
        # Create election components
        election_components = [_TestElectionComponent(i) for i in range(5)]

        # Fill in the other election components as others (instead of candidate discovery)
        self._update_others(election_components)

        # Let the third note start the election
        election_components[2].start_election()

        # The election component with the highest id should win (== 5)
        self.assertTrue(election_components[4].is_leader())

        # Close the highest election component
        election_components[4].destroy()

        # Start the election from the second highest component
        election_components[3].start_election()

        # Check if this component won
        self.assertTrue(election_components[3].is_leader())

        # Add a new highest component
        election_components.append(_TestElectionComponent(1000))
        self._update_others(election_components)

        # Let a random component start the election
        election_components[3].start_election()

        # The newly created component should win the election
        self.assertTrue(election_components[5].is_leader())


class _TestNetworkElectionComponent(NetworkElectionComponent):
    def on_elected(self, is_leader):
        pass

    def __str__(self):
        return '{} - [{}]'.format(self._id, self._others)


class TestNetworkElectionComponent(unittest.TestCase):

    def _discover(self, searching_election_component, election_components):
        """Candidate discover function"""
        self.component_list = [searching_election_component] + election_components

        # Discover the other components using the multicast request
        deferred = searching_election_component.candidate_discovery()

        # When candidates have been found, test the election procedure
        deferred.addCallback(self._test_election)

    def _stop_reactor(self, _):
        """Stop reactor callback"""
        reactor.stop()

    def _election(self, largest_id_component: ElectionComponent, other_components):
        """Do the election test"""

        # Close the normal leader
        largest_id_component.destroy()

        # Do the election
        defer = other_components[-1].start_election(candidate_discovery=False)

        # Stop the reactor after the election
        defer.addCallback(self._stop_reactor)

    def _test_election(self, _):
        # Sort the components and pass the component to the election
        sorted_components = list(
            sorted(self.component_list, key=lambda x: x.get_id()))[::-1]

        # Do the election
        self._election(sorted_components[0], sorted_components[1:])

    def test_candidate_discovery(self, amount_of_components=10):
        self.skipTest('This is not a unittest and should not be run in the unittest configuration. '
                      'As this tests requires access to the local network, this test does not work on Jenkins'
                      ' and can be seen as an \'integration\' test.')

        # Create other election components
        election_components = [_TestNetworkElectionComponent("224.0.0.1", 8005) for _ in range(amount_of_components)]

        # Create the component that starts the discovery
        searching_election_component = _TestNetworkElectionComponent("224.0.0.1", 8005)

        # Let the test run for a second
        reactor.callFromThread(self._discover, searching_election_component, election_components)

        # Let the Twisted reactor run
        reactor.run()

        # Assert if all the ids have been received
        self.assertSetEqual({i.get_id() for i in election_components + [searching_election_component]},
                            {i.get_id() for i in searching_election_component._others})

        # Sort the components
        sorted_components = list(
            sorted([searching_election_component] + election_components, key=lambda x: x.get_id()))[::-1]

        # Check if the second best one is the leader
        self.assertTrue(sorted_components[0].is_leader())


class _MockedNetworkElectionComponent(NetworkElectionComponent):
    """Mocking class for the network election component. All the transportation methods have been mocked."""

    class MockedElectionServer(MockingNetworkServer, MockingNetworkClient,
                               NetworkElectionComponent.MulticastElectionServer):
        message_type_callbacks = {
            "candidates": "_received_candidate_message",
            'election': '_retrieve_election_message'
        }

        def __init__(self, election_component):
            super(MockingNetworkServer, self).__init__()
            self.election_component = election_component
            self.replied_messages = set()

        def send_message(self, message):
            super().send_message(message)
            for other in self.election_component._others:
                other._server.datagramReceived(self.transport.value().decode('utf-8'), None)
            self.transport.clear()

    def __init__(self, x=None):
        if x is None:
            self.uuid = str(uuid.uuid4())
        else:
            self.uuid = x
        ElectionComponent.__init__(self, self.uuid)
        self._server = self.MockedElectionServer(self)
        self._discovery_time = 1
        self.i = 0

    def on_elected(self, is_leader):
        pass

    def start_election(self, candidate_discovery=False):
        self.round += 1
        _MockedNetworkElectionComponent._start_election(self, None)

    def _start_election(self, _):
        self.i += 1
        super(_MockedNetworkElectionComponent, self)._start_election(_)


class TestMockedNetworkElectionComponent(unittest.TestCase):

    def _update_others(self, election_components):
        for i, election_component in enumerate(election_components):
            election_component._others = set(election_components)

    def test_election(self):
        """Test the mocked network election component"""

        # Create election components
        election_components = list(
            reversed(
                sorted(
                    [
                        _MockedNetworkElectionComponent() for _ in range(10)
                    ], key=lambda x: x.get_id()
                )
            )
        )

        # Fill in the other election components as others (instead of candidate discovery)
        self._update_others(election_components)

        # Let the third note start the election
        election_components[2].start_election()

        # The election component with the highest id should win (== 5)
        self.assertTrue(election_components[0].is_leader())

        # Check if it really only took one round
        for election_component in election_components:
            self.assertEqual(election_component.i, 1)

        # Close the highest election component
        election_components[0].destroy()

        # Do manual candidate discovery
        self._update_others([election_component for election_component in election_components if
                             election_component.get_state() != STATES.CLOSED])

        # Start the election from the second highest component
        election_components[3].start_election()

        # Check if this component won
        self.assertTrue(election_components[1].is_leader())

        # Check if it really only took two rounds (including the previous)
        for election_component in election_components[1:]:
            self.assertEqual(election_component.i, 2)

        # Add a new highest component
        new_leader = _MockedNetworkElectionComponent('ffffffff-62cc-4738-8227-d6f87b4be188')
        election_components.append(new_leader)

        election_components = list(
            reversed(
                sorted(
                    election_components, key=lambda x: x.get_id()
                )
            )
        )

        self._update_others(election_components)

        # Let a random component start the election
        election_components[3].start_election()

        # The newly created component should win the election
        self.assertTrue(election_components[0].is_leader())

        for election_component in election_components:
            election_component._server.transport.loseConnection()


class TestTUDAElectionComponent(unittest.TestCase):
    def test_on_elected(self):
        from src import NetworkTrainUnitDataAggregator
        from src import VirtualNetworkTUDA
        from src import ConfigurationStore

        cstore = ConfigurationStore()
        cstore.set_value("tuda_port", 9090)
        cstore.set_value("tuda_mport", 9091)

        leader_tuda_el_comp = TudaElectionComponent(with_timer=False)
        leader_tuda_el_comp.on_elected(True)
        self.assertTrue(isinstance(leader_tuda_el_comp.get_tuda_object(), NetworkTrainUnitDataAggregator))
        leader_tuda_el_comp.get_tuda_object().destroy()

        cstore.set_value("tuda_port", 9092)
        cstore.set_value("tuda_mport", 9093)

        non_leader_tuda_el_comp = TudaElectionComponent(with_timer=False)
        non_leader_tuda_el_comp._leader = leader_tuda_el_comp
        non_leader_tuda_el_comp.on_elected(False)
        self.assertTrue(isinstance(non_leader_tuda_el_comp.get_tuda_object(), VirtualNetworkTUDA))
        non_leader_tuda_el_comp.get_tuda_object().destroy()



"""
This module exposes the different data containers that are used across the application and their (de)serialization logic
"""

from src.data_container.data_container import DataTypes, DataContainer, OccupancyData, ConfigUpdate, DataTypes, \
    CarriageData

__all__ = ['DataTypes', 'DataContainer', 'OccupancyData', 'ConfigUpdate', 'CarriageData']

DataTypes().add_datatype(ConfigUpdate)
DataTypes().add_datatype(OccupancyData)
DataTypes().add_datatype(CarriageData)



from twisted.internet import task

from src import ConfigurationStore, Broker, EventListener, Event
from src.election_component import NetworkElectionComponent
from src.logger import Logger
from src.train_unit_data_aggregator import InternalNetworkAPI
from src.train_unit_data_aggregator.amqp_transfer_module import AMQPTransferModule
from src.train_unit_data_aggregator.network_train_unit_data_aggregator import NetworkTrainUnitDataAggregator
from src.virtual_tuda import VirtualNetworkTUDA


class TudaElectionComponent(NetworkElectionComponent, EventListener):
    """Election component that creates a (V)TUDA when the election is done"""

    def __init__(self, with_timer=True):
        store = ConfigurationStore()
        group = store.get_value('election_mgroup')
        port = store.get_value('election_mport')

        NetworkElectionComponent.__init__(
            self, group, port, election_id=store.get_value('carriage_id'), ip=store.get_value('ip')
        )

        EventListener.__init__(self)

        self._tuda_object = None

        Broker().subscribe(self)

        self.election_deferred = None

        if with_timer:
            self.start_election_timer()

    def on_connection_failure(self):
        self.start_election()

    def on_elected(self, is_leader):
        cstore = ConfigurationStore()
        if self._tuda_object is not None:
            self._tuda_object.destroy()
        if is_leader:
            self._tuda_object = NetworkTrainUnitDataAggregator(
                cstore.get_value('tuda_ip'),
                cstore.get_value('tuda_port'),
                cstore.get_value('tuda_mgroup'),
                cstore.get_value('tuda_mport'),
                internal_api=InternalNetworkAPI(cstore.get_value('api_port')),
                transfer_module=AMQPTransferModule(
                    amqp_host=cstore.get_value('amqp_host'),
                    username=cstore.get_value('username'),
                    password=cstore.get_value('password')
                )
            )

            # Let the train unit data aggregator know which carriages are active in the train
            self._tuda_object.clear_carriages()
            discovered_carriage_ids = [carriage_id._id for carriage_id in self._others] + [self._id]
            for carriage_id in discovered_carriage_ids:
                self._tuda_object.receive_carriage_data(carriage_id)

        else:
            leader = self.get_leader()
            self._tuda_object = VirtualNetworkTUDA(
                NetworkTrainUnitDataAggregator(
                    leader.ip[0],
                    cstore.get_value('tuda_port'),
                    cstore.get_value('tuda_mgroup'),
                    cstore.get_value('tuda_mport')
                ),
                self.get_id(),
                cstore.get_value('tuda_mgroup'),
                cstore.get_value('tuda_mport')
            )

        Logger().info(
            f'Elected carriage: {self.get_leader().get_id() if not is_leader else self.get_id()}',
            location=self.__class__.__name__
        )
        Broker().publish(OnElectedEvent(self._tuda_object))

    def get_tuda_object(self):
        return self._tuda_object

    def receive_message(self, message, round_nr):
        if self.election_deferred is not None:
            self.election_deferred.reset()
        super().receive_message(message, round_nr)

    def start_election_timer(self):
        l = task.LoopingCall(self.start_election)
        l.start(ConfigurationStore().get_value('election_timer_timeout'), now=False)
        self.election_deferred = l


class OnElectedEvent(Event):
    def __init__(self, message):
        super().__init__('on_elected', message)



"""This module exposes the different implementations of the election component which orchestrates
the election on the train to select a leader."""

from src.election_component.election_component import ElectionComponent
from src.election_component.network_election_component import NetworkElectionComponent

__all__ = ['ElectionComponent', 'NetworkElectionComponent']



import datetime as dt
import json

from src.singleton import Singleton


class DataTypes(metaclass=Singleton):

    def __init__(self):
        self.types = {}
        self.reverse_types = {}
        self.curr_id = 0

    def add_datatype(self, datatype):
        if datatype not in self.reverse_types:
            self.types[self.curr_id] = datatype
            self.reverse_types[datatype] = self.curr_id
            self.curr_id += 1

    def remove_datatype(self, datatype):
        d_id = self.reverse_types[datatype]
        self.reverse_types.pop(datatype)
        self.types.pop(d_id)
        self.curr_id += 1

    def get_value(self, datatype):
        return self.reverse_types[datatype]

    def get_type(self, d_id):
        return self.types[d_id]


class DataContainer:
    """
    A general class to contain the general information of each type of data transferred through the train
    environment network such as the occupancy data extracted from the footage, the footage itself and model updates.
    """

    def __init__(self, carriage_id, timestamp=None):
        self.carriage_id = carriage_id
        self.data_type = DataTypes().get_value(self.__class__)
        self.timestamp = dt.datetime.now().astimezone().isoformat() if timestamp is None else timestamp

    def marshal(self):
        """
        Transform this python object into a JSON string such that it can be sent over the network.
        """

        return json.dumps(self.__dict__, default=self.serialize)

    def serialize(self, field):
        """
        Transform a specific field of this class object to string if json is unable to do it automatically.
        """

        return

    @staticmethod
    def unmarshal(json_data):
        """
        Transform the JSON string to a python object.
        """

        data = json.loads(json_data)
        return DataTypes().get_type(data['data_type']).to_object(data)

    @staticmethod
    def to_object(data):
        """
        Instantiate a new python object containing the given data.
        """

        return

    def __str__(self):
        return str(self.__dict__)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class OccupancyData(DataContainer):
    """
    Objects of this class represent the occupancy data from a SINGLE train carriage.
    """

    def __init__(self, carriage_id, seat_map, processed_data=None, timestamp=None):
        super().__init__(carriage_id, timestamp)
        self.seat_map = seat_map
        self.processed_data = processed_data if processed_data is not None else {}

    @staticmethod
    def to_object(data):
        return OccupancyData(
            data['carriage_id'],
            data['seat_map'],
            data['processed_data'],
            data['timestamp']
        )

    def to_API_object(self):
        obj = {"carriage": self.carriage_id, "timestamp": self.timestamp, "seatMap": self.seat_map}
        for processed_type, processed_data in self.processed_data.items():
            obj[to_camel_case(processed_type)] = processed_data
        return obj


class ConfigUpdate(DataContainer):
    """
    Objects of this class represent the updates that need to be performed on the model running in each carriage
    of the train unit. The backend emits these model updates.
    """

    def __init__(self, carriage_id, config_type, data, timestamp=None):
        super().__init__(carriage_id, timestamp)
        self.config_type = config_type
        self.data = data

    @staticmethod
    def to_object(data):
        return ConfigUpdate(data['carriage_id'], data['config_type'], data['data'])


class CarriageData(DataContainer):
    """
    Objects of this class contain the data (size, type, ...) of a carriage of the train unit.
    """

    def __init__(self, carriage_id, carriage_size, carriage_type, timestamp=None):
        super().__init__(carriage_id, timestamp)
        self.size = carriage_size
        self.type = carriage_type

    @staticmethod
    def to_object(data):
        return CarriageData(data['carriage_id'], data['size'], data['type'], data['timestamp'])


def to_camel_case(snake_case):
    """
    Convert the snake_case format to camelCase format.
    """

    first, *rest = snake_case.split('_')
    return first + ''.join(word.capitalize() for word in rest)



import json
import requests

from pika.adapters import twisted_connection
from twisted.internet import defer, protocol, reactor, task

from src.logger import Logger
from src.train_unit_data_aggregator.transfer_module import TransferModule

class AMQPTransferModule(TransferModule):
    """Transfer module which uses the AMQP transfer protocol."""

    def __init__(self, amqp_port=5672, **kwargs):
        super().__init__(**kwargs)
        self.port = amqp_port

    def attach_tuda(self, tuda):
        super().attach_tuda(tuda)
        self._listen_updates()

    def transfer_to_backend(self, data, queue):
        """Overridden method which transfers the data to the AMQP server."""
        print(data)
        cc = protocol.ClientCreator(reactor, twisted_connection.TwistedProtocolConnection, self.parameters)
        d = cc.connectTCP(self.host, self.port)
        d.addCallback(lambda p: p.ready)
        d.addCallback(self._publish, data, queue)
        d.addErrback(self._on_connection_error)
        Logger().debug("AMQP connection was set-up", location=self.__class__.__name__)

    def _on_connection_error(self, error):
        """Error callback which logs all twisted errors."""
        Logger().error(f'An error occurred setting up the AMQP connection: "{repr(error)}"',
                       location=self.__class__.__name__)

    @defer.inlineCallbacks
    def _publish(self, connection, data, queue):
        """Callback when AMQP connection is fully set-up."""
        channel = yield connection.channel()
        yield channel.queue_declare(queue=queue)
        channel.basic_publish(exchange='', routing_key=queue, body=json.dumps(data))
        Logger().debug("Sent AMQP message to queue", location=self.__class__.__name__)
        self._reset_module(channel, connection)

    def _reset_module(self, channel, connection):
        """Resets the queue & the data."""
        channel.close()
        connection.close()
        Logger().debug("AMQP connection closed and module reset", location=self.__class__.__name__)

    @defer.inlineCallbacks
    def _run(self, connection):
        channel = yield connection.channel()
        #exchange = yield channel.exchange_declare(exchange='updates')
        for carriage_id in self.tuda.get_carriages():
            queue_name = 'updates_' + str(carriage_id)
            yield channel.queue_declare(queue=queue_name, auto_delete=False, exclusive=False)
            queue_object, consumer_tag = yield channel.basic_consume(queue=queue_name)
            l = task.LoopingCall(self._read, queue_object)
            Logger().info("Start consuming " + queue_name, location=self.__class__.__name__)
            l.start(0.01)

    @defer.inlineCallbacks
    def _read(self, queue_object):
        ch, method, properties, body = yield queue_object.get()
        if body:

            Logger().info("received update", location=self.__class__.__name__)
            self.receive_config_link(json.loads(body))
        yield ch.basic_ack(delivery_tag=method.delivery_tag)

    def _listen_updates(self):
        cc = protocol.ClientCreator(reactor, twisted_connection.TwistedProtocolConnection, self.parameters)
        d = cc.connectTCP(self.host, self.port)
        d.addCallback(lambda protocol: protocol.ready)
        d.addCallback(self._run)



from src.network_daemon.multicast_network_objects import MulticastNetworkServer, MulticastNetworkClient
from src.network_daemon.tcp_network_objects import TCPClient, TCPServer, TCPFactory
from src.network_daemon.testing.MockingNetworkObject import MockingNetworkClient, MockingNetworkServer, MockingTransport

__all__ = ["MulticastNetworkClient", "MulticastNetworkServer", "MockingNetworkClient",
           "MockingNetworkServer", "MockingTransport", "TCPClient", "TCPServer", "TCPFactory"]



from src.singleton import Singleton


class ConfigurationStore(metaclass=Singleton):
    """Singleton class of a store of configuration values"""

    def __init__(self):
        self._configuration_values = {}

    def get_value(self, name):
        return self._configuration_values[name]

    def set_value(self, name, value):
        self._configuration_values[name] = value

    def __contains__(self, item):
        return self._configuration_values.__contains__(item)



"""
This module exposes the train unit data aggregator which handles the communication with the backend,
the communication of the internal API, as well as some routing logic.
"""

from src.train_unit_data_aggregator.amqp_transfer_module import AMQPTransferModule
from src.train_unit_data_aggregator.internal_api import InternalAPI
from src.train_unit_data_aggregator.internal_network_api import InternalNetworkAPI
from src.train_unit_data_aggregator.train_unit_data_aggregator import TrainUnitDataAggregator
from src.train_unit_data_aggregator.transfer_module import TransferModule

__all__ = ['TrainUnitDataAggregator', 'InternalAPI', 'InternalNetworkAPI', 'TransferModule', 'AMQPTransferModule']



import json

from twisted.internet import reactor
from twisted.internet.endpoints import TCP4ServerEndpoint

from src.data_container.data_container import DataTypes, OccupancyData
from src.logger import Logger
from src.network_daemon import TCPFactory, TCPServer, MulticastNetworkClient
from src.train_unit_data_aggregator import TrainUnitDataAggregator


class NetworkTrainUnitDataAggregator(TrainUnitDataAggregator):
    def __init__(self, ip, port, m_group, m_port, **kwargs):
        super().__init__(**kwargs)
        self.ip = ip
        self.port = port

        self.endpoint_listener = None
        self.delete_endpoint = False

        endpoint = TCP4ServerEndpoint(reactor, self.port)
        factory = TCPFactory(self.TUDAServer, self)
        endpoint.listen(factory).addCallback(self.set_endpoint_listener)

        self.multicast_server = MulticastNetworkClient(m_group, m_port)
        self.m_listener = reactor.listenMulticast(m_port, self.multicast_server, listenMultiple=True)

    def set_endpoint_listener(self, listener):
        self.endpoint_listener = listener
        if self.delete_endpoint:
            self.destroy()

    def destroy(self):
        super().destroy()
        Logger().debug('Disconnected listeners')
        if self.endpoint_listener is not None:
            self.endpoint_listener.loseConnection()
            self.endpoint_listener.connectionLost(reason=None)
        else:
            self.delete_endpoint = True
        if self.m_listener is not None:
            self.m_listener.stopListening()
        self.internal_API.destroy()

    class TUDAServer(TCPServer):
        data_type_callbacks = {
            DataTypes().get_value(OccupancyData): 'receive_occupancy_data'
        }

        def __init__(self, tuda: TrainUnitDataAggregator):
            self.tuda = tuda

        def dataReceived(self, data):
            message = json.loads(data)
            getattr(self.tuda, self.data_type_callbacks[message['data_type']])(message)

    def _global_config_update(self, model_update):
        self.multicast_server.send_message(model_update)

    def add_data_callback(self, datatype: DataTypes, callback):
        self.TUDAServer.data_type_callbacks[datatype] = callback



from collections import defaultdict

from src.logger import Logger
from src.singleton import Singleton


class Broker(metaclass=Singleton):
    """
    Broker component for pub sub system. Listeners subscribe to the broker for certain message types.
    When a certain message is published. The broker warns all the listeners that are subscribed to this message type.
    The metaclass of this class is Singleton, which allows us to call Broker(), which will always return the
    existing Broker instance or, if it doesn't exist yet, create a new one.
    """

    def __init__(self):
        """Initialization of the broker, an empty map for the message types and their listeners is created."""
        self.subscription_map = defaultdict()
        self.logger = Logger()

    def subscribe(self, listener, *args):
        """Subscribe a listener for a certain message_type.

        arguments:
        listener -- The listener that wants to subscribe to the broker
        *args -- The message types, the listener wants to subscribe to.
        """
        for message_type in args:
            if message_type not in self.subscription_map:
                self.subscription_map[message_type] = [listener]
            else:
                self.subscription_map[message_type].append(listener)
            self.logger.debug(f'Added callback for message type "{message_type}" from listener "{listener}"',
                              location=self.__class__.__name__)

    def unsubscribe(self, listener, *args):
        """Unsubscribe a listener for a certain message_type."""
        for message_type in args:
            self.subscription_map[message_type].remove(listener)
            if len(self.subscription_map[message_type]) == 0:
                self.subscription_map.pop(message_type)

    def clear_listeners(self):
        """Clear the subscription_map and thus unsubscribe all listeners"""
        self.subscription_map = defaultdict()

    def publish(self, event):
        """
        publish a message of a certain message_type to all of it's listeners.

        arguments:
        message_type -- The message type of the message to be published
        message -- The message to be published
        """
        if event.type in self.subscription_map:
            for listener in self.subscription_map[event.type]:
                listener.get_callback_map()[event.type](event.message)
                self.logger.debug(
                    f'Sent message of type "{event.type}" to "{listener}".',
                    location=self.__class__.__name__
                )



class EventListener:
    """
    Abstract Listeners class. Every listeners should implement this class.
    """

    _class_callback_map = {}

    def __init__(self):
        self._callback_map = defaultdict(list, self._class_callback_map.copy())

    def get_callback_map(self):
        return self._callback_map

    def set_message_type_callback(self, message_type, callback):
        """Set a callback for a certain message_type

        message_type -- message_type
        callback -- callback function
        """
        self._callback_map[message_type] = callback

    def __str__(self):
        return self.__class__.__name__


class Event:
    def __init__(self, message_type, message):
        self.type = message_type
        self.message = message



from src.configuration_store.configuration_store import ConfigurationStore


class ConfigurationLoader:
    """Class that handles the loading and decoding of configuration files"""

    def __init__(self, configuration_decoder, configuration_store=ConfigurationStore()):
        self.loaded_files = set()
        self.configuration_store = configuration_store
        self._configuration_decoder = configuration_decoder

    def load_config(self, filename):
        """Load the config file, decode it and add the file to the loaded files"""

        if filename in self.loaded_files:
            return
        else:
            self.loaded_files.add(filename)
            for key, value in self._configuration_decoder.decode_config(filename):
                self._set_value(key, value)

    def _set_value(self, key, value):
        """Set the value in the configuration_store"""

        self.configuration_store.set_value(key, value)


class StaticConfigurationLoader(ConfigurationLoader):
    """A configuration loader which uses the standard static ConfigurationStore"""

    def __init__(self, configuration_decoder):
        super().__init__(configuration_decoder)



import json

import pika
import requests

from src.data_container import ConfigUpdate


class TransferModule:
    """This component handles the data flow from the train unit to the remote backend and vice versa."""

    def __init__(self, amqp_host="localhost", username="", password=""):
        self.tuda = None
        self.host = amqp_host
        self.parameters = pika.ConnectionParameters(amqp_host, 5672, "/", pika.PlainCredentials(username, password))
        self.data = None
        self.queue = None

    def attach_tuda(self, tuda):
        """
        Attach the train unit data aggregator to the transfer module so that the transfer module
        can interact with it.
        """

        self.tuda = tuda

    def transfer_to_train(self, data):
        """Transfer the model update to the train unit data aggregator"""
        if self.tuda:
            self.tuda.receive_config_update(ConfigUpdate(data['carriage_id'], data['config_type'], data['data']))
        else:
            raise ValueError("Train unit data aggregator is unknown!")

    def transfer_to_backend(self, data, queue):
        """
        Transfer the given data (seat occupancy data, ...) to the backend.
        First a connection to the AMQP server is setup, next the necessary queue is declared to the channel and
        the data is published.
        """
        # TODO: remove this dirty hack
        pass

    def receive_config_link(self, body):
        r = requests.get(body['download_link'])
        body.pop('download_link')
        body['data'] = json.loads(r.text)
        self.transfer_to_train(body)



"""This module exposes the ConfigurationStore which stores configuration variables read from config files.
To read these config files, the ConfigurationLoader can be used with a certain ConfigurationDecoder which decodes
specific file formats.
"""

from src.configuration_store.configuration_decoder import ConfigurationDecoder
from src.configuration_store.configuration_loader import ConfigurationLoader, StaticConfigurationLoader
from src.configuration_store.configuration_store import ConfigurationStore
from src.configuration_store.json_configuration_decoder import JsonConfigurationDecoder

__all__ = ['ConfigurationStore', 'ConfigurationDecoder', 'JsonConfigurationDecoder', 'ConfigurationLoader',
           'StaticConfigurationLoader']



import json
import os
import random
import unittest

import pika
import requests
from twisted.internet import reactor
from twisted.web import server
from twisted.web.http import HTTPChannel

from src import EventListener, Broker, Event
from src import OccupancyData, ConfigUpdate, CarriageData
from src import MockingTransport
from src import TrainUnitDataAggregator, TransferModule
from src import AMQPTransferModule
from src import InternalNetworkAPI
from src import find_free_port


class TestPlugin(EventListener):
    def __init__(self):
        super().__init__()
        self.set_message_type_callback('update_test', self.update_configuration)
        Broker().subscribe(self, 'update_test')

    def update_configuration(self, config_update):
        path = os.path.dirname(os.path.realpath(__file__)) + '/test_config.json'
        with open(path, 'w') as file:
            file.write(json.dumps(config_update.data))


class _TestTrainUnitDataAggregator(TrainUnitDataAggregator):

    def __init__(self, transfer_module=None, internal_API=None):
        tm = transfer_module
        if transfer_module is None:
            tm = _TestTransferModule()
        super().__init__(tm, internal_API, 200)
        self.received_data = None

    def receive_config_update(self, model_update):
        self.received_data = model_update
        Broker().publish(Event('update_test', model_update))

    def _local_config_update(self, config_update, basepath_plugin=''):
        super()._local_config_update(config_update, basepath_plugin)


class _TestTransferModule(TransferModule):

    def __init__(self):
        super().__init__()
        self.received_data = None

    def transfer_to_backend(self, data, queue='queue'):
        self.received_data = data


class TestTrainUnitDataAggregator(unittest.TestCase):

    def test_transfer_to_train(self):
        test_transfer_module = _TestTransferModule()
        test_occupancy = OccupancyData(55,
                                       {'1': True, '2': False, '3': True, '4': False},
                                       {'free_seats': 2, 'fill_rate': 0.5})

        # This should throw an error as the transfer module does not know its tuda
        self.assertRaises(ValueError, test_transfer_module.transfer_to_train, test_occupancy)

    def test_receive_occupancy_data(self):
        self.skipTest("The transfermodule is now called via twisted and thus uses a reactor")
        # First let the tuda know all the carriages in the train
        test_transfer_module = _TestTransferModule()
        test_tuda = _TestTrainUnitDataAggregator(test_transfer_module)

        test_tuda.receive_carriage_data(1)
        test_tuda.receive_carriage_data(2)
        test_tuda.receive_carriage_data(3)

        test_occ_1 = OccupancyData(1,
                                   {'1': True, '2': False, '3': True, '4': False},
                                   {'free_seats': 2, 'fill_rate': 0.5})
        test_occ_2 = OccupancyData(2,
                                   {'1': False, '2': False, '3': False, '4': False},
                                   {'free_seats': 4, 'fill_rate': 0})
        test_occ_3 = OccupancyData(3,
                                   {'1': True, '2': True, '3': True, '4': True},
                                   {'free_seats': 0, 'fill_rate': 1})

        # Send the occupancies one a a time and check each time if the transfer module received the occupancies.
        # The tuda should only send the occupancies to the transfer module when it received occupancy data for
        # each carriage in the train.
        test_tuda.receive_occupancy_data(test_occ_1)
        self.assertIsNone(test_transfer_module.received_data)

        # Send occupancy for a carriage that already sent its occupancy to check if the tuda does not
        # send the occupancies to the transfer module until all occupancies are available.
        test_tuda.receive_occupancy_data(test_occ_1)
        self.assertIsNone(test_transfer_module.received_data)

        test_tuda.receive_occupancy_data(test_occ_2)
        self.assertIsNone(test_transfer_module.received_data)

        test_tuda.receive_occupancy_data(test_occ_3)
        self.assertIsNotNone(test_transfer_module.received_data)

        # Check if the tuda did reset its occupancy data map after sending all the occupancies in the map
        # to the transfer module
        self.assertEqual(test_tuda._occupancy_data_map, {1: None, 2: None, 3: None})

        test_tuda.destroy()

    def test_receive_carriage_data(self):
        test_transfer_module = _TestTransferModule()
        test_tuda = _TestTrainUnitDataAggregator(test_transfer_module)
        test_tuda._occupancy_data_map = {}

        test_tuda.receive_carriage_data(1)
        test_tuda.receive_carriage_data(2)
        test_tuda.receive_carriage_data(2)

        # The tuda should know that there are 3 carriages in the train
        expected_occupancy_map = {1: None, 2: None}
        self.assertEqual(test_tuda._occupancy_data_map, expected_occupancy_map)

        expected_occupancy_map = {1: None, 2: None, 3: None}
        test_tuda.receive_carriage_data(3)
        self.assertEqual(test_tuda._occupancy_data_map, expected_occupancy_map)

        test_tuda.clear_carriages()
        expected_occupancy_map = {test_tuda.carriage_id: None}
        self.assertEqual(test_tuda._occupancy_data_map, expected_occupancy_map)

        test_tuda.destroy()


    def test_receive_model_update(self):
        test_transfer_module = _TestTransferModule()
        test_tuda = _TestTrainUnitDataAggregator(test_transfer_module)
        test_model_update = ConfigUpdate(0, 'test_data', 'test')
        test_model_update.timestamp = 0

        # Send a dummy model update to the tuda
        test_transfer_module.transfer_to_train({"carriage_id": 0, "config_type": "test_data", "data": "test"})

        # Check if the tuda received the model update from the transfer module
        test_tuda.received_data.timestamp = 0
        self.assertEqual(test_tuda.received_data, test_model_update)

        test_tuda.destroy()

    def test_handle_network_request(self):
        api = InternalNetworkAPI(port=find_free_port())
        test_tuda = _TestTrainUnitDataAggregator(internal_API=api)
        test_occupancy = OccupancyData(55,
                                       {'1': True, '2': False, '3': True, '4': False},
                                       {'free_seats': 2, 'fill_rate': 0.5})

        # Send the occupancy to the tuda who should forward it to the internal API
        test_tuda.receive_occupancy_data(test_occupancy)
        channel = HTTPChannel()
        transport = MockingTransport()
        channel.transport = transport
        request = server.Request(channel)
        request.path = b"/occupancy"
        response = json.loads(test_tuda.internal_API.render_GET(request).decode("utf-8"))

        # Request the occupancy using a valid path: /occupancy
        response = json.loads(api.render_GET(request).decode("utf-8"))
        self.assertCountEqual(response, [test_occupancy.to_API_object()])

        path = "^/test$"
        test_tuda.internal_API.add_path(path, lambda x: None)
        self.assertTrue(path in test_tuda.internal_API.path_handlers)
        test_tuda.internal_API.remove_path(path)
        self.assertTrue(path not in test_tuda.internal_API.path_handlers)

        test_tuda.destroy()

    def test_transfer_module_AMQP(self):
        self.skipTest("This test uses a reactor and should thus be ran as a standalone test.")
        try:
            test_message = "testmessage"
            # check if connection available
            requests.get("https://www.google.com/")
            tm = AMQPTransferModule(amqp_host="193.190.127.224", username="admin", password="openspace")
            tm.transfer_to_backend(test_message, "test")
            reactor.callLater(0.5, reactor.stop)
            reactor.run()
            connection = pika.BlockingConnection(pika.ConnectionParameters("193.190.127.224", 5672, "/",
                                                                           pika.PlainCredentials("admin", "openspace")))
            channel = connection.channel()
            channel.queue_declare(queue="test")
            method_frame, header_frame, body = channel.basic_get(queue='test', no_ack=True)
            channel.close()
            connection.close()
            self.assertEqual(body.decode("utf-8")[1:-1], test_message)

        except requests.exceptions.ConnectionError:
            self.skipTest("No connection available for execution of this test")

    class _TestConnection:
        def __init__(self):
            self.open = True

        def close(self):
            self.open = False

    def test_transfer_module_AMQP_utils(self):
        from src import Logger
        import time

        r_val = random.random()
        error_string = f"Got error {r_val}"

        tm = AMQPTransferModule(amqp_port=9070)
        tm._on_connection_error(ValueError(error_string))

        time.sleep(0.5)

        with open(Logger().file_path) as f:
            lines = f.readlines()
            self.assertTrue(any([error_string in line for line in lines]))

        connection = self._TestConnection()
        channel = self._TestConnection()

        self.assertTrue(connection.open)
        self.assertTrue(channel.open)

        tm._reset_module(channel, connection)

        self.assertFalse(connection.open)
        self.assertFalse(channel.open)

    def test_update_configs(self):
        # initialisation
        dir_path = os.path.dirname(os.path.realpath(__file__))
        transfer_module = _TestTransferModule()
        tuda = _TestTrainUnitDataAggregator(transfer_module)
        transfer_module.attach_tuda(tuda)

        test_data = {
            "config_type": "test",
            "carriage_id": 200,
            "data": {
                "camera_handler": {
                    "mapping": {"1": "localhost"},
                    "camera_interval_delay": 60
                }
            }
        }
        TestPlugin()
        transfer_module.transfer_to_train(test_data)

        with open(dir_path + '/test_config.json', 'r') as file:
            config = file.read()
            self.assertEqual(config, json.dumps(test_data['data']))
        os.remove(dir_path + '/test_config.json')



import unittest

from twisted.test import proto_helpers

from src import MulticastNetworkServer, TCPClient
from src import MockingNetworkServer, MockingNetworkClient, MockingTransport


class _TestReceiver(MulticastNetworkServer):
    """Receiver to check if the datagrams have been received."""

    def __init__(self, group, port):
        super().__init__(group, port)
        self.received_datagrams = []

    def datagramReceived(self, datagram, address):
        self.received_datagrams.append(datagram)

    def startProtocol(self):
        pass


class TestMulticastNetwork(unittest.TestCase):

    def setUp(self):
        self.tr = proto_helpers.StringTransport()
        self.proto = _TestReceiver("224.0.0.1", 8005)
        self.proto.makeConnection(self.tr)

    def test_client(self, expected='test'):
        self.proto.datagramReceived(expected, ("224.0.0.1", 8005))
        self.assertListEqual(self.proto.received_datagrams, [expected])
        self.proto.datagramReceived(expected, ("224.0.0.1", 8005))
        self.assertListEqual(self.proto.received_datagrams, [expected, expected])

        self.proto.stopProtocol()
        self.tr.abortConnection()


class _TestMockupServer(MockingNetworkServer):
    """Mockup server that stores all received datagrams in a list."""

    def __init__(self):
        MockingNetworkServer.__init__(self)
        self.received_messages = []

    def datagramReceived(self, datagram, addr):
        if isinstance(datagram, bytes):
            self.received_messages += [datagram.decode('utf-8')]
        else:
            self.received_messages += [datagram]


class TestMockupObjects(unittest.TestCase):
    def test_mockup_network_object(self):
        expected = 'test'
        mock_server = _TestMockupServer()
        self.assertTrue(isinstance(mock_server.transport, MockingTransport))
        mock_client = MockingNetworkClient()
        mock_client.send_message(expected)
        mock_server.datagramReceived(mock_client.transport.value(), 'test')
        self.assertEqual(mock_client.transport.value().decode('utf-8'), expected)
        self.assertEqual(mock_server.received_messages[0], expected)
        mock_client.transport.clear()
        mock_client.send_message(str.encode(expected))
        mock_server.datagramReceived(mock_client.transport.value(), 'test')
        self.assertEqual(mock_server.received_messages[0], expected)


class TestTCPObjects(unittest.TestCase):
    def test_tcp_client(self):
        client = TCPClient()
        client.transport = MockingTransport()
        expected = 'test'
        client.send_message(expected)
        self.assertEqual(client.transport.value().decode('utf-8'), expected)
        client.transport.clear()
        client.send_message(str.encode(expected))
        self.assertEqual(client.transport.value().decode('utf-8'), expected)
        client.transport.clear()
        with self.assertRaises(ValueError):
            client.send_message(12)



from twisted.internet import reactor

from src.election_component.tuda_election_component import TudaElectionComponent
from src.logger import Logger


def setup_election_component():
    election_component = TudaElectionComponent()
    election_component.start_election()


def start_network_listeners(environment_function=None):
    """Start the reactor and let the remaining functions run on another thread"""

    Logger().info('Setup done, starting network listeners', location='SETUP')
    if environment_function is not None:
        reactor.callFromThread(environment_function)
    reactor.run()


def main():
    Logger().info('Setting up train environment', location='SETUP')

    start_network_listeners(setup_election_component())


if __name__ == '__main__':
    main()



import json
import unittest

from twisted.internet import defer

from src import DataTypes, OccupancyData, ConfigUpdate
from src import MockingNetworkTCPClient
from src import TrainUnitDataAggregator
from src import NetworkTrainUnitDataAggregator
from src import VirtualTUDA, VirtualNetworkTUDA


class _TestTUDA(TrainUnitDataAggregator):
    """TUDA class used to store occupancy data for testing purposes"""

    def __init__(self):
        super().__init__()
        self.receive_buffer = []

    def receive_occupancy_data(self, occupancy_data):
        """[Overridden] Store the received occupancy data in this class"""
        self.receive_buffer.append(occupancy_data)


class TestVirtualTUDA(unittest.TestCase):
    """Test the virtual TUDA implementation without any network logic"""

    def test_simple_forward(self):
        # Create the objects
        real_tuda = _TestTUDA()
        vtuda = VirtualTUDA(real_tuda)

        # Forward occupancy data
        vtuda.forward_occupancy_data('test_occupancy')
        self.assertTrue('test_occupancy' in real_tuda.receive_buffer)


class _TestNetworkTrainUnitDataAggregator(NetworkTrainUnitDataAggregator):
    """Class that implements the TUDA network logic and stores the received objects for testing purposes"""

    def __init__(self, ip, port, **kwargs):
        super().__init__(ip, port, '228.0.0.5', 9997, **kwargs)
        self.received_occupancy = []

    def receive_occupancy_data(self, occupancy_data):
        """[Overridden] Store the received occupancy data in this class"""
        self.received_occupancy += [occupancy_data]

class _TestNetworkVirtualTuda(VirtualNetworkTUDA):
    """Class that implements the virtual TUDA network logic and stores the received objects for testing purposes"""

    def __init__(self, leader: NetworkTrainUnitDataAggregator, name):
        super().__init__(leader, name)
        self.received_model_updates = []

    def _local_config_update(self, model_update):
        """[Overridden] Store the model updates in this class"""
        self.received_model_updates += [model_update]


class TestNetworkVirtualTUDA(unittest.TestCase):
    """
    Use the network implementations to test the responsibilities of the networked, virtual and non-virtual TUDA
    These tests should not be run under a test suite as each test creates a reactor! Each test has to be run separately
    and act more as a proof that the solution also works on a network (even though it's the localhost network).
    """

    def _send_info(self, data1, data2, real_tuda):
        # Create Vtuda's
        vtuda = VirtualNetworkTUDA(real_tuda, '127.0.0.1')
        vtuda2 = VirtualNetworkTUDA(real_tuda, '127.0.0.1')

        # Forward the occupancies
        vtuda.forward_occupancy_data(data1)
        vtuda2.forward_occupancy_data(data2)

    def test_vtuda_communication(self):
        from twisted.internet import reactor

        self.skipTest('This test uses a reactor and should thus be ran as a standalone test.')

        # Create the TUDA
        real_tuda = _TestNetworkTrainUnitDataAggregator('127.0.0.1', 9991)

        # Create the occupancy data
        data1 = {'1': False, 'data_type': DataTypes().get_value(OccupancyData)}
        data2 = {'2': False, 'data_type': DataTypes().get_value(OccupancyData)}

        # Add the callbacks to the twisted reactor and let it run
        reactor.callLater(0.1, reactor.stop)
        reactor.callFromThread(self._send_info, data1, data2, real_tuda)
        reactor.run()

        # Check if the values are correct
        # Order can be random due to the network!
        self.assertCountEqual([data1, data2], real_tuda.received_occupancy)

    def _send_info_from_tuda(self, data1, data2, real_tuda):
        real_tuda.receive_config_update(json.dumps(data1))
        real_tuda.receive_config_update(json.dumps(data2))

    def test_tuda_communication(self):
        from twisted.internet import reactor

        self.skipTest('This test uses a reactor and should thus be ran as a standalone test.')

        # Create the (virtual) TUDA's
        real_tuda = _TestNetworkTrainUnitDataAggregator('127.0.0.1', 9992)
        vtuda = _TestNetworkVirtualTuda(real_tuda, '127.0.0.1')
        vtuda2 = _TestNetworkVirtualTuda(real_tuda, '127.0.0.1')

        # Create the model update data
        data1 = {'1': False}
        data2 = {'2': False}

        # Add the callbacks to the twisted reactor and let it run
        reactor.callLater(0.5, reactor.stop)
        reactor.callFromThread(self._send_info_from_tuda, data1, data2, real_tuda)
        reactor.run()

        # Check if the values are correct
        # Order can be random due to the network!
        self.assertCountEqual(vtuda.received_model_updates, [data1, data2])
        self.assertCountEqual(vtuda2.received_model_updates, [data1, data2])


class _TestMockedNetworkTrainUnitDataAggregator(_TestNetworkTrainUnitDataAggregator):
    """
    Class that implements the TUDA network logic using the network mocking objects
    and stores the received objects for testing purposes
    """

    def __init__(self, ip, port, **kwargs):
        super().__init__(ip, port, **kwargs)
        self.received_occupancy = []
        self.server = self.TUDAServer(self)
        self.others = []

    def receive_occupancy_data(self, occupancy_data):
        """[Overridden] Store the occupancy data in this class"""
        self.received_occupancy += [occupancy_data]

    def _global_config_update(self, model_update):
        """[Overridden] Send the data to all other virtual TUDA's in the network (mimicked by a list here)"""
        for other in self.others:
            other.multicast_server.datagramReceived(model_update.marshal(), None)


class _TestMockedNetworkVTUDA(VirtualNetworkTUDA):
    """
    Class that implements the virtual TUDA network logic using the network mocking objects
    and stores the received objects for testing purposes
    """

    def __init__(self, leader: NetworkTrainUnitDataAggregator, name):
        super().__init__(leader, name, '228.0.0.5', 9997)
        self.received_model_updates = []

    def open_tcp(self):
        """[Overridden] Open connection to a mocked up TCP client"""
        d = defer.Deferred()
        d.callback(MockingNetworkTCPClient())
        return d

    def _send_message(self, endpoint, message):
        """[Overridden] Send a message to a given endpoint (which is the leader)"""
        self.leader.server.dataReceived(json.dumps(message))

    def _local_config_update(self, config_update, basepath_plugin='../plugin'):
        """[Overridden] Store the model update in this class"""
        self.received_model_updates += [ConfigUpdate(
            config_update['carriage_id'],
            config_update['config_type'],
            config_update['data'],
            timestamp=config_update['timestamp']
        )]



class TestMockedNetworkVirtualTUDA(unittest.TestCase):
    """Use the mocked implementations to test the responsibilities of the networked, virtual and non-virtual TUDA"""

    def test_vtuda_communication(self):
        # Create the (virtual) TUDA's
        real_tuda = _TestMockedNetworkTrainUnitDataAggregator('127.0.0.1', 9993)
        vtuda = _TestMockedNetworkVTUDA(real_tuda, '127.0.0.1')
        vtuda2 = _TestMockedNetworkVTUDA(real_tuda, '127.0.0.1')

        # Create the occupancy data
        data1 = {'1': False, 'data_type': DataTypes().get_value(OccupancyData)}
        data2 = {'2': False, 'data_type': DataTypes().get_value(OccupancyData)}

        # Forward the first occupancy and check the result
        vtuda.forward_occupancy_data(data1)
        self.assertListEqual([data1], real_tuda.received_occupancy)

        # Forward the second occupancy and check the result
        vtuda2.forward_occupancy_data(data2)
        self.assertListEqual([data1, data2], real_tuda.received_occupancy)

        # Try another random number of forwards and check
        for i in range(10):
            vtuda2.forward_occupancy_data(data2)
        self.assertListEqual([data1, data2] + ([data2] * 10), real_tuda.received_occupancy)

    def test_tuda_to_vtuda_communication(self):
        # Create the (virtual) TUDA's
        real_tuda = _TestMockedNetworkTrainUnitDataAggregator('127.0.0.1', 9994)
        vtuda = _TestMockedNetworkVTUDA(real_tuda, '127.0.0.1')
        vtuda2 = _TestMockedNetworkVTUDA(real_tuda, '127.0.0.1')

        # Create the model update data
        data1 = ConfigUpdate(0, 'test_data', 'test')
        data2 = ConfigUpdate(0, 'test_data', 'tast')

        # Add the virtual TUDA's to the real TUDA (only for testing purposes)
        # This mimics a multicast network
        real_tuda.others = [vtuda, vtuda2]

        # Receive and forward both model updates
        real_tuda.receive_config_update(data1)
        real_tuda.receive_config_update(data2)

        # Check the values in each virtual TUDA
        self.assertListEqual([data1, data2], vtuda.received_model_updates)
        self.assertListEqual([data1, data2], vtuda2.received_model_updates)



import os
import unittest

import cv2

from plugin.camera_handler import Frame, CameraFootage
from src import DataContainer, OccupancyData, ConfigUpdate, CarriageData


class TestDataContainer(unittest.TestCase):

    def test_create_occupancy_container(self):
        carriage_id = 5
        seat_map = {'1': True, '2': False, '3': True, '4': False}
        free_seats, fill_rate = 2, 0.5
        test_occupancy = OccupancyData(carriage_id, seat_map, {'free_seats': free_seats, 'fill_rate': fill_rate})

        # Check if the container has all the right data
        self.assertEqual(test_occupancy.carriage_id, carriage_id)
        self.assertEqual(test_occupancy.seat_map, seat_map)
        self.assertEqual(test_occupancy.processed_data['free_seats'], free_seats)
        self.assertEqual(test_occupancy.processed_data['fill_rate'], fill_rate)
        self.assertIsNotNone(test_occupancy.timestamp)

    def test_create_camera_footage_container(self, img_file='testfiles/images/SampleImage.png'):
        from plugin.camera_handler import CameraFootage
        carriage_id, camera_id = 5, 13
        codec = 'codec'

        dir_path = os.path.dirname(os.path.realpath(__file__))
        img_file = dir_path + '/' + img_file

        # Test the 3 different ways to create a frame
        frame_3 = Frame(camera_id, codec, img_file)
        frame_2 = Frame(camera_id, codec, frame_3.image)
        test_camera_footage = CameraFootage(carriage_id, [frame_2])
        test_camera_footage.add_frame(frame_3)

        # Check if the container has all the right data
        self.assertEqual(test_camera_footage.carriage_id, carriage_id)
        self.assertEqual(len(test_camera_footage.frames), 2)

        self.assertEqual(test_camera_footage.frames[0], frame_2)
        self.assertEqual(test_camera_footage.frames[1], frame_3)
        self.assertEqual(test_camera_footage.frames[0], test_camera_footage.frames[1])

        self.assertIsNotNone(test_camera_footage.timestamp)

    def test_create_config_update_container(self):
        carriage_id = 5
        test_type = "test_type"
        model_update = {"test": "test"}
        test_model_update = ConfigUpdate(carriage_id, test_type, model_update)

        # Check if the container has all the right data
        self.assertEqual(test_model_update.carriage_id, carriage_id)
        self.assertEqual(test_model_update.data, model_update)
        self.assertIsNotNone(test_model_update.timestamp)

    def test_create_carriage_data_container(self):
        carriage_id, carriage_size, carriage_type = 5, 80, 1
        test_carriage_data = CarriageData(carriage_id, carriage_size, carriage_type)

        # Check if the container has all the right data
        self.assertEqual(test_carriage_data.carriage_id, carriage_id)
        self.assertEqual(test_carriage_data.size, carriage_size)
        self.assertEqual(test_carriage_data.type, carriage_type)
        self.assertIsNotNone(test_carriage_data.timestamp)

    def test_marshal_unmarshal_occupancy_container(self):
        carriage_id = 5
        seat_map = {'1': True, '2': False, '3': True, '4': False}
        free_seats, fill_rate = 2, 0.5
        test_occupancy = OccupancyData(carriage_id, seat_map, free_seats, fill_rate)

        # Transform object to json so it can be sent over the network
        json = test_occupancy.marshal()

        # Transform the json back to the original object
        test_occupancy_2 = DataContainer.unmarshal(json)

        # Transform object to json so it can be sent over the network
        json = test_occupancy_2.marshal()

        # Transform the json back to the original object
        test_occupancy_3 = DataContainer.unmarshal(json)

        self.assertEqual(test_occupancy, test_occupancy_3)

    def test_marshal_unmarshal_camera_footage_container(self, file='./test/testfiles/images/SampleImage.png'):
        carriage_id, camera_id = 5, 13
        codec = 'codec'

        image_bytes = cv2.imread(file)

        frame = Frame(camera_id, codec, image_bytes)
        test_camera_footage = CameraFootage(carriage_id, [frame])

        # Transform object to json so it can be sent over the network
        json = test_camera_footage.marshal()

        # Transform the json back to the original object
        test_camera_footage_2 = DataContainer.unmarshal(json)

        # Transform object to json so it can be sent over the network
        json = test_camera_footage_2.marshal()

        # Transform the json back to the original object
        test_camera_footage_3 = DataContainer.unmarshal(json)

        self.assertEqual(test_camera_footage, test_camera_footage_3)

    def test_marshal_unmarshal_model_update_container(self):
        carriage_id = 5
        config_type = "test_type"
        model_update = {"test": "test"}
        test_model_update = ConfigUpdate(carriage_id, config_type, model_update)

        # Transform object to json so it can be sent over the network
        json = test_model_update.marshal()

        # Transform the json back to the original object
        test_model_update_2 = DataContainer.unmarshal(json)

        # Transform object to json so it can be sent over the network
        json = test_model_update_2.marshal()

        # Transform the json back to the original object
        test_model_update_3 = DataContainer.unmarshal(json)

        test_model_update_3.timestamp = 0
        test_model_update.timestamp = 0
        self.assertEqual(test_model_update, test_model_update_3)

    def test_marshal_unmarshal_carriage_data_container(self):
        carriage_id, carriage_size, carriage_type = 5, 80, 1
        test_carriage_data = CarriageData(carriage_id, carriage_size, carriage_type)

        # Transform object to json so it can be sent over the network

        json = test_carriage_data.marshal()

        # Transform the json back to the original object
        test_carriage_data_2 = DataContainer.unmarshal(json)

        # Transform object to json so it can be sent over the network
        json = test_carriage_data_2.marshal()

        # Transform the json back to the original object
        test_carriage_data_3 = DataContainer.unmarshal(json)

        self.assertEqual(test_carriage_data, test_carriage_data_3)



"""
Logger util module.
This module wraps the logging library of python, but handles the writing of logging files in a more standard way.
There is also only one Logger object as it uses the singleton metaclass. Due to this, the first invocation of a logger
object should set the correct path of the log file.
"""

from src.logger.logger import Logger

__all__ = ["Logger"]



from collections import defaultdict
from enum import Enum

from src.logger import Logger


class STATES(Enum):
    """An enum of the different states an election component can be in."""

    IDLE = 0  # The component is idle and can accept messages (default state)
    SENT_ELECTION = 1  # The component sent an election message
    CLOSED = 2  # The component is closed and will not reply to messages
    LEADER = 3  # The component is the leader


class ElectionComponent:
    """A component that handles the election procedure on the train."""

    def __init__(self, id, ip=None):
        self._id = id
        self._others = set()
        self._state = STATES.IDLE
        self._answers = defaultdict(list)
        self._leader = None
        self._time_out = 2  # TODO: make this part of the configuration
        self.round = 0
        self.election_rounds_participated = set()
        self.ip = ip

    def __eq__(self, other):
        return self.get_id() == other.get_id()

    def __hash__(self):
        return hash(self.get_id())

    def candidate_discovery(self):
        """Discover other candidates for the election"""
        Logger().debug('Starting candidate discovery', location=self.__class__.__name__)
        candidates = [i for i in self._others if i.get_state() != STATES.CLOSED]
        Logger().debug('Finished candidate discovery', location=self.__class__.__name__)
        return candidates

    def receive_message(self, message, round_nr):
        self._on_receive_message(message, round_nr)

    def _on_receive_message(self, message, round_nr):
        """Do something when a message is received.
        When the update_leader message has been received and the component has not yet participated in the
        current round, it discovers the candidates in the network and restarts the election without broadcasting."""
        if message == 'update_leader' and round_nr not in self.election_rounds_participated and self.get_state() != STATES.CLOSED:
            self.election_rounds_participated.add(round_nr)
            self.round = round_nr
            self._start_election(True)

    def send_message(self, source, message):
        """Send a message to a given source"""
        source.receive_message(message, self.round)

    def broadcast_message(self, message):
        """Broadcast a message to all other candidates"""
        for candidate in self._others:
            self.send_message(candidate, message)

    def _start_election(self, candidate_discovery):
        self._leader = max(
            self.candidate_discovery() + [self] if candidate_discovery else self._others.union(set(self)),
            key=lambda x: x.get_id()
        )
        is_leader = self._leader.get_id() == self.get_id()
        self.election_rounds_participated.add(self.round)
        self._on_elected(is_leader)
        self.broadcast_message('update_leader')

    def start_election(self, candidate_discovery=True):
        """Start the election procedure by sending an election message to all higher ID'ed candidates"""
        self.round += 1
        self._start_election(candidate_discovery=candidate_discovery)

    def _on_elected(self, is_leader):
        """Logic to be ran when a leader has been elected"""
        if is_leader:
            self._state = STATES.LEADER
        self.on_elected(is_leader)

    def on_elected(self, is_leader):
        """Callback that will be called when a leader has been elected"""
        raise NotImplementedError

    def get_leader(self):
        return self._leader

    def is_leader(self):
        return self.get_state() == STATES.LEADER

    def get_state(self):
        return self._state

    def get_id(self):
        return self._id

    def destroy(self):
        """Destroy this election component. This will set its state to closed."""
        self._state = STATES.CLOSED



import unittest

from src import EventListener, Broker, Event


class _TestSingleListener(EventListener):
    def __init__(self):
        super().__init__()
        self.fill_listener_map()
        self.test_value = 0

    def fill_listener_map(self):
        self.get_callback_map()['test'] = self.test_callback

    def test_callback(self, message):
        self.test_value = message


class _TestMultipleListener(EventListener):
    _class_callback_map = {}

    def __init__(self):
        super().__init__()
        self.fill_listener_map()
        self.test_value = 0

    def fill_listener_map(self):
        self.get_callback_map()['test'] = self.test_callback
        self.get_callback_map()['test_double'] = self.test_double_callback
        self.get_callback_map()['test_triple'] = self.test_triple_callback

    def test_callback(self, message):
        self.test_value = message

    def test_double_callback(self, message):
        self.test_value = 2 * message

    def test_triple_callback(self, message):
        self.test_value = 3 * message


class BrokerTests(unittest.TestCase):
    def test_pub_sub_single(self):
        broker = Broker()
        listener = _TestSingleListener()

        # Subscribe & Publish
        broker.subscribe(listener, 'test')
        broker.publish(Event('test', 5))

        # Check if any message was received
        self.assertEqual(listener.test_value, 5)

    def test_pub_sub_multiple(self):
        broker = Broker()
        listener = _TestMultipleListener()

        # Subscribe
        broker.subscribe(listener, 'test', 'test_double', 'test_triple')

        # Test if correct callbacks are applied
        broker.publish(Event('test_triple', 5))
        self.assertEqual(listener.test_value, 15)
        broker.publish(Event('test', 5))
        self.assertEqual(listener.test_value, 5)
        broker.publish(Event('test_double', 5))
        self.assertEqual(listener.test_value, 10)

    def test_pub_sub_multiple_listeners(self):
        broker = Broker()
        listener1 = _TestMultipleListener()
        listener2 = _TestSingleListener()
        listener3 = _TestSingleListener()

        # Subscribe & publish
        broker.subscribe(listener1, 'test', 'test_double')
        broker.subscribe(listener2, 'test')
        broker.subscribe(listener3, 'test')
        broker.publish(Event('test', 5))

        # Test if all listeners receive message
        self.assertEqual(listener1.test_value, 5)
        self.assertEqual(listener2.test_value, 5)
        self.assertEqual(listener3.test_value, 5)

        broker.publish(Event('test_double', 5))
        self.assertEqual(listener1.test_value, 10)
        self.assertEqual(listener2.test_value, 5)
        self.assertEqual(listener3.test_value, 5)

    def test_subscribe_unsubscribe(self):
        broker = Broker()
        listener = _TestMultipleListener()

        # Subscribe
        broker.subscribe(listener, 'test', 'test_double', 'test_triple')

        # Test if correct callbacks are applied
        broker.publish(Event('test_triple', 5))
        self.assertEqual(listener.test_value, 15)
        broker.publish(Event('test', 5))
        self.assertEqual(listener.test_value, 5)
        broker.publish(Event('test_double', 5))
        self.assertEqual(listener.test_value, 10)

        broker.unsubscribe(listener, 'test')

        # Publish some messages and check if listener doesn't receive the 'test' message
        broker.publish(Event('test_triple', 5))
        self.assertEqual(listener.test_value, 15)
        broker.publish(Event('test', 5))
        self.assertEqual(listener.test_value, 15)
        broker.publish(Event('test_double', 5))
        self.assertEqual(listener.test_value, 10)

        broker.clear_listeners()

        # Publish some messages and check if listener doesn't receive them
        broker.publish(Event('test_triple', 5))
        self.assertEqual(listener.test_value, 10)
        broker.publish(Event('test', 5))
        self.assertEqual(listener.test_value, 10)
        broker.publish(Event('test_double', 5))
        self.assertEqual(listener.test_value, 10)



import json

from src.configuration_store.configuration_decoder import ConfigurationDecoder


class JsonConfigurationDecoder(ConfigurationDecoder):
    """Json configuration decoder"""

    def decode_config(self, filename):
        with open(filename) as json_data:
            for key, value in json.load(json_data).items():
                yield key, value



import json
import re

from twisted.internet import reactor
from twisted.web import resource, server
from twisted.web.resource import NoResource

from src.train_unit_data_aggregator import InternalAPI


class InternalNetworkAPI(resource.Resource, InternalAPI):
    """
    An object of this class represents a server listening for HTTP requests received by the network of the train.
    """

    isLeaf = True

    def __init__(self, port):
        resource.Resource.__init__(self)
        InternalAPI.__init__(self)

        self.path_handlers = {"^/occupancy$": self.handle_GET_occupancy}
        self.server = server.Site(self)
        self.port = port

        self.http_listener = reactor.listenTCP(self.port, self.server)

    def destroy(self):
        if self.http_listener is not None:
            self.http_listener.loseConnection()
            self.http_listener.connectionLost(reason=None)

    def render_GET(self, request):
        handler = self._resolve_path_handler(request.path.decode("utf-8"))
        if handler is not None:
            response = handler()
        else:
            return NoResource()
        request.responseHeaders.addRawHeader(b"content-type", b"application/json")
        return response.encode("utf-8")

    def _resolve_path_handler(self, path):
        """
        Check if the requested path is defined for this server.
        If the path is known then return the correct handler to resolve the request.
        """

        for path_match, handler in self.path_handlers.items():
            if re.search(path_match, path):
                return handler
        return None

    def handle_GET_occupancy(self):
        # /occupancy
        return json.dumps([occ for occ in self.occupancy_data_map.values()])

    def add_path(self, path, callback):
        self.path_handlers[path] = callback

    def remove_path(self, path):
        self.path_handlers.pop(path)




import json

from twisted.internet import reactor
from twisted.internet.endpoints import TCP4ClientEndpoint, connectProtocol

from src.network_daemon import TCPClient, MulticastNetworkServer
from src.train_unit_data_aggregator.network_train_unit_data_aggregator import NetworkTrainUnitDataAggregator
from src.virtual_tuda.virtual_tuda import VirtualTUDA


class VirtualNetworkTUDA(VirtualTUDA):
    """Virtual train unit data aggregator which connects to the real one using the network"""

    class VTudaMulticastNetworkServer(MulticastNetworkServer):
        """Multicast server to receive information from the train unit data aggregator"""

        def __init__(self, vtuda, group, port):
            super().__init__(group, port)
            self.vtuda = vtuda

        def datagramReceived(self, datagram, addr):
            data = json.loads(datagram)
            self.vtuda.receive_config_update(data)

    def __init__(self, leader: NetworkTrainUnitDataAggregator, name, m_group, m_port):
        super().__init__(leader)
        self.name = name
        self.multicast_server = self.VTudaMulticastNetworkServer(self, m_group, m_port)
        self.m_listener = reactor.listenMulticast(m_port, self.multicast_server, listenMultiple=True)

    def destroy(self):
        if self.m_listener is not None:
            self.m_listener.stopListening()
        self.leader.destroy()

    def open_tcp(self):
        point = TCP4ClientEndpoint(reactor, self.leader.ip, self.leader.port)
        d = connectProtocol(point, TCPClient())
        return d

    def _send_message(self, endpoint, message):
        endpoint.send_message(json.dumps(message))

    def forward_occupancy_data(self, occupancy_data):
        """Forward the occupancy data to the real TUDA."""
        d = self.open_tcp()
        d.addCallback(lambda x: self._send_message(x, occupancy_data))
        return d

    def receive_config_update(self, model_update):
        super(VirtualNetworkTUDA, self).receive_config_update(model_update)



import os
import time
import unittest

from src import Logger


class TestLogger(unittest.TestCase):
    def test_logging_methods(self):
        dir_path = os.path.dirname(os.path.realpath(__file__))

        Logger(f'{dir_path}/test.log')
        Logger().info('info', location='TEST')
        Logger().debug('debug', location='TEST')
        Logger().warning('warning', location='TEST')
        Logger().error('error', location='TEST')

        log_types = ['INFO', 'DEBUG', 'WARNING', 'ERROR']

        time.sleep(2)

        with open(f'{dir_path}/test.log', mode='w+') as fp:
            lines = fp.readlines()

        for i, line in enumerate(lines):
            self.assertListEqual(line.strip().split(' - ')[2:], [log_types[i], f'[TEST]: {log_types[i].lower()}'])

        os.remove(f'{dir_path}/test.log')



import importlib
import os
import traceback
from pathlib import Path

from src.broker import Broker, EventListener, Event
from src.configuration_store import ConfigurationStore, ConfigurationLoader, JsonConfigurationDecoder
from src.train_env.main import main
from src.singleton import Singleton

name = "openspacetrainenv"

__all__ = ["Broker", "ConfigurationStore", "ConfigurationLoader", "ConfigurationDecoder", "JsonConfigurationDecoder",
           "EventListener", "Event", "DataTypes", "DataContainer", "OccupancyData", "Singleton"]

dir_path = os.path.dirname(os.path.realpath(__file__))
current_exec_path = Path().absolute()

store = ConfigurationStore()
loader = ConfigurationLoader(JsonConfigurationDecoder(), store)
loader.load_config(f'{current_exec_path}/main_config.json')

try:
    exec(
        compile(open(f'{current_exec_path}/settings.py', "rb").read(), f'{current_exec_path}/settings.py', 'exec'),
        globals(),
        locals()
    )
    for k, v in globals()['KWARGS'].items():
        store.set_value(k, v)

except Exception as e:
    traceback.print_tb(e.__traceback__)


def init_as_main():
    load_plugins_from_path(dir_path + '/../plugin')
    for plugin in globals()['PLUGINS']:
        load_plugins_from_pip(plugin)
    main()


def load_plugins_from_pip(package):
    if package not in globals():
        try:
            importlib.import_module(package)
        except ImportError:
            print(f'Invalid package: The package with name "{package}" was not installed on the system. Please make '
                  f'sure it is installed, either using pip or present in the plugin folder.')
        finally:
            globals()[package] = importlib.import_module(package)
    else:
        print(f'Using the plugin from the plugin folder instead of its pip package.')


def load_plugins_from_path(path, check_globals=False):
    import sys
    tree = os.listdir(path)

    sys.path.insert(0, path)
    for plugin in tree:
        if check_globals and plugin in globals():
            load_plugin(os.path.join(path, plugin), plugin)
        else:
            print(f'Already imported plugin with name: {plugin}')


def load_plugin(path, plugin_name):
    if os.path.isdir(path):
        plugin_obj = importlib.import_module('plugin.{}'.format(plugin_name))
        globals()[plugin_name] = plugin_obj


if __name__ == '__main__':
    init_as_main()



from twisted.internet.protocol import Factory, Protocol


class TCPClient(Protocol):
    """Client that can be used for TCP connections"""

    def send_message(self, message):
        if isinstance(message, str):
            self.transport.write(str.encode(message))
        elif isinstance(message, bytes):
            self.transport.write(message)
        else:
            raise ValueError('Invalid message, should be bytes or str.')


class TCPServer(Protocol):
    """TCP server which listens to a given port"""

    def dataReceived(self, data):
        raise NotImplementedError


class TCPFactory(Factory):
    """Factory which builds a certain TCP server connection"""

    def __init__(self, class_name, *args):
        self.class_name = class_name
        self.args = args

    def buildProtocol(self, addr):
        return self.class_name(*self.args)



"""This module contains the virtual train unit data aggregator which forwards its requests to the real
train unit data aggregator, as well as accepts messages from its leader TUDA.
"""

from src.virtual_tuda.virtual_network_tuda import VirtualNetworkTUDA
from src.virtual_tuda.virtual_tuda import VirtualTUDA

__all__ = ["VirtualTUDA", "VirtualNetworkTUDA"]



import json
import unittest

from twisted.web import server
from twisted.web._responses import NOT_FOUND
from twisted.web.http import HTTPChannel

from src import OccupancyData
from src import MockingTransport
from src import InternalNetworkAPI
from src import find_free_port


class TestInternalNetworkAPI(unittest.TestCase):

    def test_handle_bad_path_request(self):
        api = InternalNetworkAPI(port=find_free_port())

        # Send the internal network API a request containing an path that does not exist and
        # is therefore unable to handle the request

        # Create Request
        channel = HTTPChannel()
        transport = MockingTransport()
        channel.transport = transport
        request = server.Request(channel)
        request.path = b"/this/is/a/bad/request"

        response = api.render_GET(request)

        self.assertEqual(response.code, NOT_FOUND)

    def test_handle_good_request(self):
        api = InternalNetworkAPI(port=find_free_port())

        # Create Request
        channel = HTTPChannel()
        transport = MockingTransport()
        channel.transport = transport
        request = server.Request(channel)
        request.path = b"/occupancy"

        response = json.loads(api.render_GET(request).decode("utf-8"))

        self.assertEqual(response, [])

        test_occ_1 = OccupancyData(55,
                                   {'1': True, '2': False, '3': True, '4': False},
                                   {'free_seats': 2, 'fill_rate': 0.5})
        api.receive_occupancy_data(test_occ_1)
        test_occ_2 = OccupancyData(56,
                                   {'1': True, '2': False, '3': True, '4': True},
                                   {'free_seats': 1, 'fill_rate': 0.75})
        api.receive_occupancy_data(test_occ_2)

        # Request the occupancy using a valid path: /occupancy
        response = json.loads(api.render_GET(request).decode("utf-8"))
        self.assertCountEqual(response, [test_occ_1.to_API_object(), test_occ_2.to_API_object()])

        # Add another carriage with the same ID as the first carriage in order to test if the data stored
        # in the internal API is correctly updated
        test_occ_3 = OccupancyData(55,
                                         {'1': True, '2': True, '3': True, '4': True},
                                         {'free_seats': 0, 'fill_rate': 1})
        api.receive_occupancy_data(test_occ_3)

        response = json.loads(api.render_GET(request).decode("utf-8"))
        self.assertCountEqual(response, [test_occ_3.to_API_object(), test_occ_2.to_API_object()])

        api.destroy()



import os
import unittest

from src import ConfigurationStore, JsonConfigurationDecoder, ConfigurationLoader

dirname = os.path.dirname(__file__)


class TestConfigurationStore(unittest.TestCase):
    def test_store(self):
        configuration_store = ConfigurationStore()

        # Fill in the values
        for i in range(0, 100):
            configuration_store.set_value(i, i)

        # Check the values
        for i in range(0, 100):
            self.assertEqual(configuration_store.get_value(i), i)

    def test_singleton_store(self):
        configuration_store_1 = ConfigurationStore()
        configuration_store_2 = ConfigurationStore()

        # Check if both stores are equal and no new store has been created.
        self.assertEqual(configuration_store_1, configuration_store_2)

    def test_types(self):
        configuration_store = ConfigurationStore()
        types = [set, dict, list, int, float, str]

        # Fill in the values
        for i, value in enumerate(types):
            configuration_store.set_value(i, value())

        # Check the values
        for i, value in enumerate(types):
            self.assertEqual(configuration_store.get_value(i), value())


class TestJsonConfigurationDecoder(unittest.TestCase):
    def test_config_decoder(self):
        decoded = JsonConfigurationDecoder().decode_config(f'{dirname}/../test/testfiles/json/test.json')

        # Map from json
        result_map = {k: v for k, v in decoded}

        # Map we should get
        default_map = {"a": 1, "b": 2, "c": 3}

        # Check if both are the same
        self.assertEqual(result_map, default_map)


class TestJsonConfigurationLoader(unittest.TestCase):
    def test_static_config_loader(self):
        json_file = f'{dirname}/../test/testfiles/json/test.json'

        # We use our own created ConfigurationStore to not share the same store across tests
        config_loader = ConfigurationLoader(JsonConfigurationDecoder())

        # Load the config
        config_loader.load_config(json_file)

        # Check if the loaded files contain the last loaded file
        self.assertTrue(json_file in config_loader.loaded_files)

        # Map we should get
        default_map = {"a": 1, "b": 2, "c": 3}

        # Check if the values are present in the configuration store and if they are equal
        for k, v in default_map.items():
            self.assertEqual(config_loader.configuration_store.get_value(k), v)



import json
import uuid

from twisted.internet import reactor, defer

from src.election_component.election_component import ElectionComponent, STATES
from src.logger import Logger
from src.network_daemon import MulticastNetworkServer, MulticastNetworkClient


class NetworkElectionComponent(ElectionComponent):
    """Election component that uses multicast messages to discover other election components"""

    class MulticastElectionServer(MulticastNetworkServer, MulticastNetworkClient):
        """Multicast election server which takes care of sending messages as well as
        retrieving multicast messages.
        """

        message_type_callbacks = {
            "candidates": "_received_candidate_message",
            'election': '_retrieve_election_message'
        }

        def __init__(self, group, port, election_component):
            super().__init__(group, port)
            self.election_component = election_component
            self.replied_messages = set()

        def _received_candidate_message(self, message, address):
            """Receive and process a candidate message"""
            self.election_component._add_candidate(message["id"], address)

            if message["message_id"] not in self.replied_messages \
                    and self.election_component.get_state() != STATES.CLOSED:
                self.replied_messages.add(message["message_id"])
                response = {
                    "id": self.election_component.get_id(),
                    "message_id": message["message_id"],
                    "message_type": message["message_type"]
                }
                self.send_message(json.dumps(response))

        def _retrieve_election_message(self, message, _):
            """Receive and process an election message"""
            self.election_component.receive_message(message['message'], message['round'])

        def datagramReceived(self, datagram, address):
            """Overriden method from twisted. Reroutes the information to the correct method"""
            message = json.loads(datagram)
            getattr(self, self.message_type_callbacks[message["message_type"]])(message, address)

    def __init__(self, group, port, election_id=None, **kwargs):
        super().__init__(election_id if election_id is not None else str(uuid.uuid4()), **kwargs)
        self.group = group
        self.port = port
        self._server = self.MulticastElectionServer(self.group, self.port, self)
        reactor.listenMulticast(self.port, self._server, listenMultiple=True)
        self._discovery_time = 1

    def _discover(self):
        """Send a multicast message across the network (to the same multicast group)
        Do note that this method takes some time (one to multiple seconds) to come to
        a stable situation.
        """

        message = {"id": self.get_id(), "message_id": str(uuid.uuid4()), "message_type": "candidates"}
        self._server.replied_messages.add(message["message_id"])
        self._server.send_message(json.dumps(message))

    def _add_candidate(self, candidate, ip):
        """Add an election candidate"""
        self._others.add(ElectionComponent(candidate, ip=ip))

    def candidate_discovery(self):
        """Returns the list of retrieved candidates, using a set discovery time."""
        Logger().debug('Starting candidate discovery', location=self.__class__.__name__)
        deferred = defer.Deferred()
        self._discover()
        reactor.callLater(self._discovery_time, deferred.callback, None)
        return deferred

    def broadcast_message(self, message):
        self.send_message(None, message)

    def _start_election(self, _):
        """Do the local election. Check all others and pick the highest.
        Also broadcast this message so the others do the same.
        """
        self._leader = max(
            self._others,
            key=lambda x: x.get_id()
        )
        is_leader = self._leader.get_id() == self.get_id()
        self.election_rounds_participated.add(self.round)
        self._on_elected(is_leader)
        self.broadcast_message('update_leader')

    def start_election(self, candidate_discovery=True):
        """Handle callbacks for election. No real election logic is done here."""
        deferred = defer.Deferred()
        self.round += 1
        if candidate_discovery:
            candidate_defer = self.candidate_discovery()
            candidate_defer.addCallback(self._start_election)
            candidate_defer.addCallback(
                lambda x: Logger().debug('Finished candidate discovery', location=self.__class__.__name__))
        else:
            self._start_election(None)

        reactor.callLater(
            self._time_out + (0 if not candidate_discovery else self._discovery_time),
            deferred.callback,
            None
        )
        return deferred

    def receive_message(self, message, round_nr):
        """Receive a message from the network."""
        self._on_receive_message(message, round_nr)

    def send_message(self, other, message):
        """Send a message over the network to the given candidate"""
        message = {
            'message': message,
            'source': self.get_id(),
            'message_type': 'election',
            'round': self.round
        }
        self._server.send_message(json.dumps(message))



import setuptools

print(setuptools.find_packages())

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="openspacetrainenv",
    version="0.0.9",
    author="OpenSpace",
    description="The OpenSpace train environment",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.ugent.be/OpenSpace-DP/train_env",
    packages=['train_env'],
    package_dir={'': 'src'},
    classifiers=[
        "Programming Language :: Python :: 3"
    ],
)



from src.train_unit_data_aggregator import TrainUnitDataAggregator


class VirtualTUDA(TrainUnitDataAggregator):

    def __init__(self, leader: TrainUnitDataAggregator):
        super().__init__()
        self.leader = leader

    def forward_occupancy_data(self, occupancy_data):
        """Override this to use different forwarding techniques"""
        self.leader.receive_occupancy_data(occupancy_data)

    def receive_occupancy_data(self, occupancy_data):
        self.forward_occupancy_data(occupancy_data)

    def receive_config_update(self, model_update):
        self._local_config_update(model_update)



import json

from twisted.internet import reactor

from src import ConfigurationStore
from src.broker.broker import Broker, EventListener, Event
from src.logger import Logger
from src.train_unit_data_aggregator.internal_api import InternalAPI
from src.train_unit_data_aggregator.transfer_module import TransferModule

BASEPATH_PLUGINS = 'plugin'

class TrainUnitDataAggregator(EventListener):
    """
    This component is used as a central gathering point for all data from each carriage.
    The gathered data is used by the internal API and the transfer module. This component also forwards
    data received from the transfer module to the carriages.
    """

    def __init__(self, transfer_module=None, internal_api=None, carriage_id=None):
        super().__init__()
        self.internal_API = internal_api if internal_api is not None else InternalAPI()
        self.transfer_module = transfer_module if transfer_module is not None else TransferModule()
        self.carriage_id = ConfigurationStore().get_value('carriage_id') if carriage_id is None else carriage_id
        self._occupancy_data_map = {self.carriage_id: None}

        self.broker = Broker()
        self.add_broker_subscription("processed_occupancy", self.receive_occupancy_data)

        if self.transfer_module.tuda is None:
            self.transfer_module.attach_tuda(self)

    def destroy(self):
        """Method that handles cleanup of listeners amongst other things."""
        self.broker.unsubscribe(self, *self.get_callback_map().keys())

    def add_broker_subscription(self, message_type, callback):
        self.broker.subscribe(self, message_type)
        self.set_message_type_callback(message_type, callback)

    def receive_occupancy_data(self, occupancy_data):
        """"
        Receive and store seat occupancy data from a train carriage and forward it to the internal API.
        Send the occupancies to the backend whenever there is occupancy data available for each carriage in the train.
        """
        Logger().info('Received value @ TUDA', location=self.__class__.__name__)
        self.internal_API.receive_occupancy_data(occupancy_data)
        self._occupancy_data_map[occupancy_data.carriage_id] = occupancy_data
        self._forward_occupancy_data()

    def _forward_occupancy_data(self):
        """
        Send the entire occupancy data of the entire train to the backend when the occupancy from each carriage in
        the train is known and reset the occupancy data map after sending the occupancies.
        """

        missing_occupancies = sum([occ is None for occ in self._occupancy_data_map.values()])
        if missing_occupancies == 0:
            occupancies = [occ.to_API_object() for occ in self._occupancy_data_map.values()]
            # Send and reset
            reactor.callFromThread(self.transfer_module.transfer_to_backend, {"occupancy": occupancies}, 'occupancy')
            self._occupancy_data_map = {carriage_id: None for carriage_id in self._occupancy_data_map.keys()}

    def receive_carriage_data(self, carriage_id):
        """
        Receive the carriage ID from a train carriage and store this ID in the occupancy data map in order to known
        all carriages in this train. The tuda should only receive carriage data when an election
        procedure is taking place (= when all carriages of the train are being discovered).
        """

        if carriage_id not in self._occupancy_data_map.keys():
            self._occupancy_data_map[carriage_id] = None

    def _local_config_update(self, config_update, basepath_plugin = '../plugin'):
        """Send the model update to the local broker."""
        if config_update.config_type is not 'main':
            Broker().publish(Event('update_' + config_update.config_type, config_update))
        else:
                with open('main_config.json', 'w') as file:
                    file.write(json.dumps(config_update.data))

    def _global_config_update(self, config_update):
        """Send the model to the virtual train unit data aggregators on the local network"""

        pass

    def receive_config_update(self, config_update):
        """Receive a model update from the backend and forward it to each train carriage connected to the network."""

        if config_update.carriage_id == self.carriage_id:
            self._local_config_update(config_update)
        else:
            self._global_config_update(config_update)

    def clear_carriages(self):
        """Delete the known carriage IDs and this should only be done when a new election is finished."""
        self._occupancy_data_map = {self.carriage_id: None}

    def get_carriages(self):
        return self._occupancy_data_map.keys()



import math
import os
import unittest
from time import sleep

import cv2
import numpy

from plugin.camera_handler import CameraHandler, CodecListener, IntervalCameraHandler, Transcoder, \
    ImageCameraHandler, VideoCameraHandler, MultiVideoCameraHandler, MultiVideoTranscoder, CameraFootage, StreamHandler, \
    Frame
from src import Broker
from src import DataTypes

dirname = os.path.dirname(__file__)


class TestCodecListener(CodecListener):
    __test__ = False

    def on_publish(self, footage):
        self.received_footage += [footage]

    def __init__(self, codec):
        super().__init__(self.on_publish, codec)
        self.received_footage = []


class TestIntervalCameraHandler(IntervalCameraHandler):
    __test__ = False

    def __init__(self, delay, codec):
        super().__init__(delay)
        self.codec = codec

    def _retrieve_footage(self):
        return TestFootage(self.codec)


class TestFootage(CameraFootage):
    __test__ = False

    def __init__(self, codec):
        super().__init__(None, [Frame(None, codec, f'{dirname}/../test/testfiles/images/SampleImage.png')])


class CameraHandlerTests(unittest.TestCase):
    def setUp(self):
        DataTypes().add_datatype(TestFootage)

    def test_pub_sub_simple(self):
        codec = 'codec'
        broker = Broker()
        cam_handler = CameraHandler(0)
        listener = TestCodecListener(codec)
        transcoder = Transcoder(codec, lambda x: x)
        transcoder.add_footage_source('camera_handler')

        # Subscribe & Publish
        broker.subscribe(listener, codec)
        cam_handler.publish(TestFootage(codec))

        # Check if any footage was received
        self.assertTrue(listener.received_footage is not None)

    def test_pub_sub_multiple(self):
        broker = Broker()
        codec = 'codec'
        cam_handler = CameraHandler(0)
        transcoder = Transcoder(codec, lambda x: x)
        transcoder.add_footage_source('camera_handler')
        listeners = [TestCodecListener(codec) for i in range(0, 10)]

        # Subscribe all the listeners & Publish
        for listener in listeners:
            broker.subscribe(listener, codec)
        cam_handler.publish(TestFootage(codec))

        # Check if all listeners received the footage
        for listener in listeners:
            self.assertEqual(len(listener.received_footage), 1)

    @staticmethod
    def simple_transform(footage):
        footage.codec = 'y'
        return footage

    def test_transform(self):
        broker = Broker()
        codec = 'y'
        cam_handler = CameraHandler(0)
        transcoder = Transcoder(codec, self.simple_transform)
        transcoder.add_footage_source('camera_handler')
        listener = TestCodecListener(codec)

        # Subscribe & Publish
        broker.subscribe(listener, codec)
        cam_handler.publish(TestFootage('x'))

        # Check if any footage was received
        self.assertEqual(len(listener.received_footage), 1)

        # Check if the transform has been applied successfully
        self.assertEqual(listener.received_footage[0].codec, 'y')


class ImageCameraHandlerTests(unittest.TestCase):
    def setUp(self):
        DataTypes().add_datatype(TestFootage)

    def image_publish(self, delay=1, handler_delay=0.5, img_file='../test/testfiles/images/SampleImage.png'):
        # Create an Image Camera Handler with a random object
        broker = Broker()
        broker.clear_listeners()
        codec = 'numpy'
        transcoder = Transcoder(codec, lambda x: x)
        transcoder.add_footage_source('camera_handler')
        handler = ImageCameraHandler(handler_delay, f'{dirname}/{img_file}', 0)
        listener = TestCodecListener(codec)
        broker.subscribe(listener, codec)

        # Start the handler
        handler.start()
        self.assertNotEqual(handler.thread, None)

        # Sleep for a given delay
        sleep(delay)

        # Stop the handler
        handler.stop()
        self.assertEqual(handler.thread, None)

        # Check if we received enough footage objects
        self.assertAlmostEqual(len(listener.received_footage), math.floor(delay / handler_delay), delta=1)

        # Check if the retrieved frame equals the frame we sent to the CameraHandler
        self.assertTrue(
            numpy.array_equal(cv2.imread(f'{dirname}/{img_file}'), listener.received_footage[0].frames[0].image))

    def test_all_image_formats(self):
        formats = ['jpg', 'png']
        for codec in formats:
            self.image_publish(img_file=f'../test/testfiles/images/SampleImage.' + codec)


class StreamHandlerTests(unittest.TestCase):
    def frame_publish(self, delay=1, handler_delay=0.5, video_file='../test/testfiles/videos/SampleVideo.mp4'):
        broker = Broker()
        broker.clear_listeners()
        codec = 'numpy'
        # Create an Image Camera Handler with a random object
        handler = StreamHandler(handler_delay, f'{dirname}/{video_file}', 0)
        listener = TestCodecListener(codec)
        transcoder = Transcoder(codec, lambda x: x)
        transcoder.add_footage_source('camera_handler')
        broker.subscribe(listener, codec)

        # Start the handler
        handler.start()
        self.assertNotEqual(handler.thread, None)

        # Sleep for a given delay
        sleep(delay)

        # Stop the handler
        handler.stop()
        self.assertEqual(handler.thread, None)

        # Check if we received enough footage objects, higher delta, because of the opening and closing of VideoCapture
        self.assertAlmostEqual(len(listener.received_footage), math.floor(delay / handler_delay), delta=2)

    def test_all_video_formats(self):
        formats = ['flv', 'mkv', 'mp4']
        for codec in formats:
            self.frame_publish(video_file=f'../test/testfiles/videos/SampleVideo.' + codec)


class VideoCameraHandlerTests(unittest.TestCase):
    def setUp(self):
        DataTypes().add_datatype(TestFootage)

    def video_publish(self, delay=1, handler_delay=0.5, video_file='../test/testfiles/videos/SampleVideo.mp4'):
        self.skipTest(
            "This test was created before the introduction of config files and streamhandler is not relevant anymore")
        broker = Broker()
        broker.clear_listeners()
        codec = 'numpy'
        # Create an Image Camera Handler with a random object
        handler = VideoCameraHandler(handler_delay, f'{dirname}/{video_file}', 0)
        listener = TestCodecListener(codec)
        transcoder = Transcoder(codec, lambda x: x)
        transcoder.add_footage_source('camera_handler')
        broker.subscribe(listener, codec)

        # Start the handler
        handler.start()
        self.assertNotEqual(handler.thread, None)

        # Sleep for a given delay
        sleep(delay)

        # Stop the handler
        handler.stop()
        self.assertEqual(handler.thread, None)

        # Check if we received enough footage objects
        self.assertAlmostEqual(len(listener.received_footage), math.floor(delay / handler_delay), delta=1)

        # Check if the retrieved frame equals the frame we sent to the CameraHandler
        video_capture = cv2.VideoCapture(f'{dirname}/{video_file}')
        for footage_object in listener.received_footage:
            success, capture = video_capture.read()

            self.assertTrue(numpy.array_equal(capture, footage_object.frames[0].image))

    def test_all_video_formats(self):
        formats = ['flv', 'mkv', 'mp4']
        for codec in formats:
            self.video_publish(video_file=f'../test/testfiles/videos/SampleVideo.{codec}')

    def test_multi_video_publish(self, delay=4, handler_delay=2):
        self.skipTest(
            "This test was created before the introduction of config files and stream handler is not relevant anymore")
        broker = Broker()
        broker.clear_listeners()
        codec = 'numpy'
        file_map = {
            'mp4': f'{dirname}/../test/testfiles/videos/SampleVideo.mp4',
            'flv': f'{dirname}/../test/testfiles/videos/SampleVideo.flv',
            'mkv': f'{dirname}/../test/testfiles/videos/SampleVideo.mkv'
        }

        handler = MultiVideoCameraHandler(handler_delay, file_map, 0)
        listener = TestCodecListener(codec)
        transcoder = MultiVideoTranscoder(codec, lambda x: x)
        transcoder.add_footage_source('camera_handler')
        broker.subscribe(listener, codec)

        # Start the handler
        handler.start()
        self.assertNotEqual(handler.thread, None)

        # Sleep for a given delay
        sleep(delay)

        # Stop the handler
        handler.stop()
        self.assertEqual(handler.thread, None)

        # Check if we received enough footage objects
        self.assertAlmostEqual(len(listener.received_footage), math.floor(delay / handler_delay), delta=1)

        video_capture_map = {key: cv2.VideoCapture(value) for key, value in file_map.items()}

        # Check if the retrieved frame equals the frame we sent to the CameraHandler
        for footage_object in listener.received_footage:
            for key, value in file_map.items():
                success, capture = video_capture_map[key].read()
                self.assertTrue(numpy.array_equal(capture, footage_object[key].frames[0].image))



"""This module contains the broker class, used to send events around the train environment. The broker uses a Pub-Sub architecture.
All communication passes through this broker, which is a singleton, so only one is present on the entire environment.
Plugins should override the Event class so that they can publish these events to other components that implement the EventListener interface.
"""

from src.broker.broker import Broker, EventListener, Event

__all__ = ['Broker', 'EventListener', 'Event']



import logging
from pathlib import Path

from src.configuration_store import ConfigurationStore
from src.singleton import Singleton


class Logger(metaclass=Singleton):
    """
    Logger which logs to the given file path and wraps the logging library methods.
    The logging_levels are the same as the logging library in descending levels:

    CRITICAL
    ERROR
    WARNING
    INFO
    DEBUG
    NOTSET

    All levels above the given level will be logged.
    For example if one chooses 'DEBUG' as a log level in the config,
    DEBUG, INFO, WARNING, ERROR & CRITICAL will be logged.
    """

    def __init__(self, file_path=None):
        self._logging_level = None

        current_exec_path = Path().absolute()
        self.file_path = f'{current_exec_path}/{file_path}' if file_path is not None \
            else f'{current_exec_path}/{ConfigurationStore().get_value("logging_path")}'
        if 'logging_level' in ConfigurationStore():
            self._logging_level = ConfigurationStore().get_value('logging_level')
            logging.basicConfig(level=getattr(logging, self._logging_level), filename=self.file_path, filemode='w+',
                                format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                                datefmt='%d-%m-%y %H:%M:%S')

    def log(self, log_level, message, *args, location='', **kwargs):
        if self._logging_level is not None:
            prefix = f'[{location}]:' if location is not '' else ''
            return logging.log(log_level, f'{prefix} {message}', *args, **kwargs)
        else:
            pass

    def debug(self, *arg, **kwargs):
        self.log(logging.DEBUG, *arg, **kwargs)

    def info(self, *arg, **kwargs):
        self.log(logging.INFO, *arg, **kwargs)

    def warning(self, *arg, **kwargs):
        self.log(logging.WARNING, *arg, **kwargs)

    def error(self, *arg, **kwargs):
        self.log(logging.ERROR, *arg, **kwargs)
